
import coinlib.dataWorker_pb2 as statsModel
import coinlib.dataWorker_pb2_grpc as stats
import threading
import time
from coinlib.helper import log
import pandas as pd
import queue
import grpc
import simplejson as json
import traceback
import zlib
import binascii
from coinlib.Registrar import Registrar
from coinlib.InfluxDatabase import InfluxDatabase
from chipmunkdb.ChipmunkDb import ChipmunkDb

class WorkerJobProcess:
    def __init__(self, workerJob, factory):
        if factory is not None:
            self.workerChannel = factory.createChannel()
        self.registrar = Registrar()
        self.chipmunkDb = ChipmunkDb(self.registrar.chipmunkdb)
        if factory is not None:
            self.stub = stats.DataWorkerStub(self.workerChannel)
        self.workerJob = workerJob
        self.stopped = True
        self.currentWorkerJobData = []
        self.factory = factory
        self.dataFrame = pd.DataFrame()
        return None

    def closeChannel(self):
        self.workerChannel.close()

    def generateRawData(self, result_data):

        result_data_json = str(json.dumps(result_data, ignore_nan=True)).encode('utf8')
        result_data_json_compressed = zlib.compress(bytes(result_data_json))

        return result_data_json_compressed

    def getDataFrameColumn(self, key, column_name, _type="base"):
        try:
            if column_name+":open" in self.dataFrame.columns:
                dfCopy = self.dataFrame[[column_name+":open", column_name+":high",
                                        column_name+":low", column_name+":close", column_name+":volume"]].copy()
                dfCopy["open"] = dfCopy[column_name+":open"]
                dfCopy["high"] = dfCopy[column_name + ":high"]
                dfCopy["low"] = dfCopy[column_name + ":low"]
                dfCopy["close"] = dfCopy[column_name + ":close"]
                dfCopy["volume"] = dfCopy[column_name + ":volume"]

                #self.dataFrame[key + ":open"] = self.dataFrame[column_name + ":open"]
                #self.dataFrame[key + ":open"] = self.dataFrame[column_name + ":high"]
                #self.dataFrame[key + ":low"] = self.dataFrame[column_name + ":low"]
                #self.dataFrame[key + ":close"] = self.dataFrame[column_name + ":close"]
                #self.dataFrame[key + ":volume"] = self.dataFrame[column_name + ":volume"]

                return dfCopy
            elif column_name+":y" in self.dataFrame.columns:
                return self.dataFrame[column_name + ":y"].copy()

        except Exception as e:
            log.error("Assert error")
            pass

        return self.dataFrame[column_name]

    def extractInputsFromDataFrameAndInsertInDataFrame(self, inputs):
        raw_inputs = {}
        for key in inputs:
            try:
                inputVal = inputs[key]
                type = inputVal["type"]
                value = inputVal["value"]
                if type == "symbol":
                    val = value
                    if isinstance(val, dict):
                        val = value["id"] if "id" in value else value["value"]
                    raw_inputs[key] = self.getDataFrameColumn(key, val, _type="ohlc")
                elif type == "dataInput" or type == "any":
                    val = value
                    if isinstance(val, dict):
                        val = value["id"] if "id" in value else value["value"]
                    raw_inputs[key] = self.getDataFrameColumn(key, val, _type="ohlc")
                elif type == "feature":
                    val = value
                    if isinstance(val, dict):
                        val = value["id"] if "id" in value else value["value"]
                    raw_inputs[key] = self.getDataFrameColumn(key, val)
                elif type == "any":
                    raw_inputs[key] = value
                    self.dataFrame[key] = value
                elif type == "number":
                    raw_inputs[key] = value
                    self.dataFrame[key] = value
                elif type == "int":
                    raw_inputs[key] = value
                    self.dataFrame[key] = value
                elif type == "float":
                    raw_inputs[key] = value
                    self.dataFrame[key] = value
                else:
                    raw_inputs[key] = value
                    self.dataFrame[key] = value
            except Exception as e:
                tb = traceback.format_exc()
                log.error("Problem on extracting data", e)
                pass

        return raw_inputs

    def initialize(self):
        pass
    
    def getDf(self):
        return self.dataFrame
    
    def getChannel(self):
        return self.workerChannel
        
    def getWorkerJobDataFrame(self, workerJob):
        df = pd.DataFrame()
        if hasattr(self, 'chartConfigData'):

            configData = self.chartConfigData
            start = time.time()
            if configData.chart_prefix:
                df = self.chipmunkDb.collection_as_pandas(configData.workspace_id, columns=[],
                                                          domain=configData.chart_prefix)
            else:
                df = self.chipmunkDb.collection_as_pandas(configData.workspace_id, columns=[])
            end = time.time()
            log.info("Downloading DataFrame from chipmunk ", end - start)

            if configData.chart_prefix != "" and configData.chart_prefix is not None:
                df.columns = df.columns.str.replace('^'+configData.chart_prefix + ".", "", regex=True)

                if "symbol." not in df.columns:
                    # we need to copy open, high, low, close
                    key = "symbol"
                    df[key + ":open"] = df["main:open"]
                    df[key + ":high"] = df["main:high"]
                    df[key + ":low"] = df["main:low"]
                    df[key + ":close"] = df["main:close"]
                    df[key + ":volume"] = df["main:volume"]

        return df

        # currentWorkerJobData = b''
        #for dataBlock in self.stub.GetWorkerJobData(workerJob):
        #    currentWorkerJobData = currentWorkerJobData + bytes(dataBlock.data)

        #if len(currentWorkerJobData) > 0:
        #    decompressed_data = zlib.decompress(currentWorkerJobData, 15 + 32)
        #    df = pd.DataFrame(json.loads(decompressed_data))
        #    df['Datetime'] = pd.to_datetime(df['datetime'], unit='s')
        #    df = df.set_index(['Datetime'])
        #else:
        #    df = pd.DataFrame()

        #return df
    
    def runProcess(self):
        try:
            self.run()

        except Exception as e:
            log.error(e)
            self.onErrorProcess(e)
            return

        log.info("Finished")
        self.onFinishedProcess()
    
    def run(self):
        pass

    def onErrorProcess(self, e):
        self.stop()
        self.factory.onWorkerJobProcessError(self, e)

    def onFinishedProcess(self):
        self.stop()
        self.factory.onWorkerJobProcessFinished(self)
        
    def stop(self):
        self.stopped = True
        #self.listenThread.join()
        
    def onBeforeDownloadData(self):
        pass

    def reloadDataFrame(self):
        self.dataFrame = self.getWorkerJobDataFrame(self.workerJob)
        return True
    
    def downloadAndRunprocess(self):
        try:
            self.onBeforeDownloadData()

            self.initialize()
            self.dataFrame = self.getWorkerJobDataFrame(self.workerJob)
            if (self.stopped == False):
                self.stop()
                self.stopped = True

            self.listenThread = threading.Thread(target=self.runProcess, daemon=True)
            self.listenThread.start()
        except Exception as e:
            self.onErrorProcess(e)
            return False
            
        return True

    def startProcessWithDataFrame(self, dataFrame):
        try:
            self.onBeforeDownloadData()

            self.initialize()
            self.dataFrame = dataFrame
            if (self.stopped == False):
                self.stop()
                self.stopped = True

            self.listenThread = threading.Thread(target=self.runProcess, daemon=True)
            self.listenThread.start()

        except Exception as e:
            self.onErrorProcess(e)
            return False

        return True

    def startProcess(self):
        
        downloadThread = threading.Thread(target=self.downloadAndRunprocess, daemon=True)
        downloadThread.start()

        pass

    def setConfig(self, configuration):
        pass
    
    


import os
import sys
import inspect
from typing import TypedDict

from coinlib.broker.Broker import BrokerAssetType, BrokerContractType
from coinlib.brokerWorker.BrokerFactory import BrokerFactory
from coinlib.feature.FeatureFactory import FeatureFactory
from coinlib.logics.LogicLoader import LogicFactory
from coinlib.helper import get_current_kernel
import coinlib.dataWorker_pb2 as statsModel
import coinlib.dataWorker_pb2_grpc as stats
from grpc._cython.cygrpc import CompressionAlgorithm
from grpc._cython.cygrpc import CompressionLevel
import grpc
from coinlib.helper import in_ipynb
from coinlib.notification.NotificationFactory import NotificationFactory
from coinlib.statistics.StatisticsMethodFactory import StatisticsMethodFactory
from coinlib.ChartsFactory import ChartsFactory
from coinlib.statistics.StatisticsRuleFactory import StatisticsRuleFactory

from coinlib.Registrar import Registrar
import simplejson as json

from coinlib.DataWorker import WorkerJobListener
from coinlib.symbols.SymbolFactory import SymbolFactory

class BrokerLoginInfo(TypedDict):
    apikey: str
    secret: str

class BrokerTestLoginData(TypedDict):
    exchange_id: str
    paperMode: bool
    loginInfo: BrokerLoginInfo
    assetType: BrokerAssetType
    contractType: BrokerContractType


class Simulator:
    def __init__(self):
        self._isConnected = False
        self._registrar = Registrar()
        self.workerJobs = None

        return None

    def connectSimulator(self):
        if self.__isLiveStage() == False:
            self.channel = self.__createChannel()
            self.stub = stats.SimulatorStub(self.channel)

    def connect(self):
        self._isConnected = True
        if not self._registrar.isRegistered:
            self.__registerAsWorker()

    def __isPluginFile(self):
        if self._registrar.currentPluginLoading is not None:
            callstack = inspect.stack()
            for call in callstack:
                if self._registrar.currentPluginLoading in call.filename:
                    return True
        return False

    def __canNotUseSimulatorInCurrentFile(self):
        return self.__isLiveStage() or self.__isPluginFile()

    def __isLiveStage(self):
        return self._registrar.isLiveEnvironment()

    def __registerAsWorker(self):
        if self.__canNotUseSimulatorInCurrentFile():
            return False
        if not self._registrar.isRegistered:
            self._registrar.isRegistered = True

            self._registrar.statsRuleFactory = StatisticsRuleFactory()
            self._registrar.statsMethodFactory = StatisticsMethodFactory()
            self._registrar.chartsFactory = ChartsFactory()
            self._registrar.logicFactory = LogicFactory()
            self._registrar.brokerFactory = BrokerFactory()
            self._registrar.symbolFactory = SymbolFactory()
            self._registrar.featureFactory = FeatureFactory()
            self._registrar.notificationFactory = NotificationFactory()

            self.workerJobs = WorkerJobListener(simulator=True)
            self.workerJobs.start()

        return True

    def __createChannel(self):
        chan_ops = [('grpc.max_receive_message_length', 1000 * 1024 * 1024),
                    ('grpc.default_compression_algorithm', CompressionAlgorithm.gzip),
                    ('grpc.grpc.default_compression_level', CompressionLevel.high)]
        return grpc.insecure_channel(self._registrar.get_coinlib_backend_grpc(), options=chan_ops,
                                     compression=grpc.Compression.Gzip)

    def plot(self, plotly_fig):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerAsWorker()

        json = plotly_fig.to_json()

        iframe_src = 'http://' + self._registrar.iframe_host + '/simulator/plotly/' + get_current_kernel(
            self.workerJobs.workerId)
        if in_ipynb():

            self.__displayCoinChartFrame(iframe_src, iframe_width=str(1140), message='{type: "jupyter", plotly:'+json+', command: "reloadPlotlyChartData"}')


        return True

    def testBroker(self, broker_id: str, broker_login_data: BrokerTestLoginData):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        print("We are testing your symbol broker - but this can take 4-5 minutes - we´ll send you some infos when finished.");

        brokerInputs = statsModel.BrokerSimulatorRequest()
        brokerInputs.brokerId = broker_id
        brokerInputs.workerId = self._registrar.worker_id
        brokerInputs.paperMode = broker_login_data["paperMode"]

        brokerInfoData = statsModel.BrokerSimulatorBrokerData()
        brokerInfoData.exchange_id = broker_login_data["exchange_id"]

        brokerInfoData.assetType = "COIN"
        if "assetType" in broker_login_data:
            if broker_login_data["assetType"] == BrokerAssetType.COIN:
                brokerInfoData.assetType = "COIN"
            if broker_login_data["assetType"] == BrokerAssetType.STOCK:
                brokerInfoData.assetType = "STOCK"
            if broker_login_data["assetType"] == BrokerAssetType.FOREX:
                brokerInfoData.assetType = "FOREX"

        brokerInfoData.contractType = "spot"
        if "contractType" in broker_login_data:
            if broker_login_data["contractType"] == BrokerContractType.FUTURE:
                brokerInfoData.contractType = "future"
            if broker_login_data["contractType"] == BrokerContractType.SPOT:
                brokerInfoData.contractType = "spot"
            if broker_login_data["contractType"] == BrokerContractType.MARGIN:
                brokerInfoData.contractType = "margin"
            if broker_login_data["contractType"] == BrokerContractType.OPTION:
                brokerInfoData.contractType = "option"

        brokerLoginInfo = statsModel.BrokerSimulatorBrokerLoginData()
        brokerLoginInfo.secret = broker_login_data["loginInfo"]["secret"]
        brokerLoginInfo.apikey = broker_login_data["loginInfo"]["apikey"]

        brokerInfoData.loginInfo.CopyFrom(brokerLoginInfo)

        brokerInputs.brokerInfo.CopyFrom(brokerInfoData)

        info = self.stub.simulateBroker(brokerInputs)
        exit = False
        if info.success:
            print("Your symbol Broker does work, read message after!")

            print(info.simulatorResponseText+"\r\n")

            iframe_src = 'http://localhost:3000/simulator/broker/' + info.workspaceId + "/" + get_current_kernel(
                self.workerJobs.workerId) + "/" + info.sessionId + "/" + info.targetdepotId
            if in_ipynb():

                self.__displayCoinChartFrame(iframe_src,
                                             iframe_width=str(1140),
                                             messageData=str(json.dumps({"workspaceId": info.workspaceId},
                                                                        ignore_nan=True)))
            else:
                # we are in anything of?????
                # printing iframe:
                print("Broker works - here is your Testframe: " + iframe_src)

        else:
            exit = True
            print("ERROR! Your symbol Broker does not work, read message after!")
            print(info.simulatorResponseText)


        return info

    def simulateFeature(self, identifier, options=None):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        print("We are testing your Feature Class. When finished we will report the data to you.")
        print("  => Please beware: When you have a lake Feature - we will start it and then please trigger some data inputs. ")
        print("  => This process will finish when any data is inserted into the data storage and we will return the info to you")

        notificationInputs = statsModel.FeatureSimulatorRequest()
        notificationInputs.identifier = identifier
        notificationInputs.workerId = self._registrar.worker_id
        notificationInputs.options = str(json.dumps(options, ignore_nan=True))

        info = self.stub.simulateFeature(notificationInputs)
        exit = False
        if info.success:
            print("Your Feature system works fine!!")
        else:
            exit = True
            print("ERROR! Your Feature service does not work, read message after!")

        print(info.simulatorResponseText)

        return info

    def testNotification(self, identifier, channels, inputs):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        print("We are testing your Notification and sending two messages to it. Please click on the second one to finish the test")

        notificationInputs = statsModel.NotificationSimulatorRequest()
        notificationInputs.identifier = identifier
        notificationInputs.channels = str(json.dumps(channels, ignore_nan=True))
        notificationInputs.workerId = self._registrar.worker_id
        notificationInputs.inputs = str(json.dumps(inputs, ignore_nan=True))

        info = self.stub.simulateNotification(notificationInputs)
        exit = False
        if info.success:
            print("Your Notification system works fine!!")
        else:
            exit = True
            print("ERROR! Your Notification service does not work, read message after!")

        print(info.simulatorResponseText)

        return info



    def testSymbolBroker(self, symbolbroker_id, testSymbol1, testSymbol2, options):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        print("We are testing your symbol broker - but this can take 4-5 minutes - we´ll send you some infos when finished.");

        symbolBrokerInputs = statsModel.SymbolBrokerSimulatorRequest()
        symbolBrokerInputs.symbolBrokerId = symbolbroker_id
        symbolBrokerInputs.workerId = self._registrar.worker_id
        symbolBrokerInputs.testSymbol1 = testSymbol1
        symbolBrokerInputs.testSymbol2 = testSymbol2
        symbolBrokerInputs.options = str(json.dumps(options, ignore_nan=True))

        info = self.stub.simulateSymbolBroker(symbolBrokerInputs)
        exit = False
        if info.success:
            print("Your symbol Broker does work, read message after!")
        else:
            exit = True
            print("ERROR! Your symbol Broker does not work, read message after!")

        print(info.simulatorResponseText)

        return info

    def __registerSimulator(self):
        self.connectSimulator()

    def statsMethod(self, methods, name="simulatorDefaultWorkspace"):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        simulatorChartConfig = statsModel.SimulatorMethodCallChartConfig()
        simulatorChartConfig.kernel_id = get_current_kernel(self.workerJobs.workerId)
        simulatorChartConfig.simulatorName = name
        simulatorChartConfig.workerId = self.workerJobs.workerId

        for method in methods:
            meth = statsModel.SimulatorMethodCall()
            meth.method = method["method"]
            meth.params = str(json.dumps(method["params"], ignore_nan=True))

            simulatorChartConfig.methods.append(meth)

        info = self.stub.simulateStatisticsMethod(simulatorChartConfig)

        iframe_src = 'http://localhost:3000/simulator/statsMethod/' + info.workspaceId + "/" + get_current_kernel(self.workerJobs.workerId)
        if in_ipynb():


            self.__displayCoinChartFrame(iframe_src,
                                         iframe_width=str(1140),
                                         messageData=str(json.dumps({"methods": methods,
                                                                     "workspaceId": info.workspaceId}, ignore_nan=True)))
        else:
            # we are in anything of?????
            # printing iframe:
            print("Your Testframe: "+iframe_src)

        return True

    def statsRule(self, name="simulatorDefaultWorkspace"):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        simulatorChartConfig = statsModel.SimulatorStatisticChartConfig()
        simulatorChartConfig.kernel_id = get_current_kernel(self.workerJobs.workerId)
        simulatorChartConfig.simulatorName = name

        info = self.stub.simulateStatisticsRule(simulatorChartConfig)
        iframe_src = 'http://localhost:3000/simulator/stats/' + info.workspaceId + "/" + get_current_kernel(self.workerJobs.workerId)
        if in_ipynb():

            self.__displayCoinChartFrame(iframe_src,
                                         messageData=str(json.dumps({"workspaceId": info.workspaceId}, ignore_nan=True)),
                                         iframe_width=str(1140))


        return True

    def __displayCoinChartFrame(self, iframe_src, iframe_width = str(980), iframe_height = str(360), messageData='', message=''):
        from IPython.display import Javascript
        from IPython.core.display import display
        postmessage = message
        if (messageData == ''):
            messageData = "{}"

        if message == "":
            postmessage = '{type: "jupyter", kernel_id: "'+get_current_kernel(self.workerJobs.workerId)+'", command: "reloadChartDataFromNotebook", data: '+messageData+'}'
        display(Javascript(' const currentNotebookPanel = document.querySelector(".jp-NotebookPanel:not(.p-mod-hidden)");'\
                           ' if (currentNotebookPanel.querySelector(".coinchartoutput") != null) '\
                           ' { '\
                           '    currentNotebookPanel.querySelector(".coinchartoutput").querySelector("iframe").contentWindow.postMessage('+postmessage+', "*"); '\
                           ' } '\
                           ' else '\
                           ' { '\
                           '    const elem = document.createElement("div"); '\
                           '    elem.innerHTML = "<iframe src=\'' + iframe_src + '\' border=0 width=\'' + iframe_width + '\' height=\'' + iframe_height + '\' />";'\
                           '    elem.style = "position: absolute;width:' + iframe_width + 'px;height:' + iframe_height + 'px;bottom: 0px;opacity: 0.9;z-index: 9999;"; '\
                           '    elem.className = "coinchartoutput";'\
                           '    currentNotebookPanel.appendChild(elem); '\
                           ' } '))

        return True

    def chart(self, broker, symbol, timeframe, indicators):

        if self.__canNotUseSimulatorInCurrentFile():
            return False
        self.__registerSimulator()

        simulatorChartConfig = statsModel.SimulatorChartConfig()
        simulatorChartConfig.kernel_id = get_current_kernel(self.workerJobs.workerId)
        simulatorChartConfig.broker = broker
        simulatorChartConfig.symbol = symbol
        simulatorChartConfig.timeframe = timeframe
        simulatorChartConfig.workerId = self.workerJobs.workerId
        indilist = []
        for indi in indicators:
            indicator = statsModel.SimulatorChartConfigIndicator()
            indicator.feature = indi["feature"]
            indicator.subfeature = indi["subfeature"]
            if "chartIndex" in indi:
                indicator.chartIndex = indi["chartIndex"]

            for key in indi["inputs"]:
                indicator.inputs[key] = str(indi["inputs"][key])

            simulatorChartConfig.indicators.append(indicator)

        self.stub.simulateChart(simulatorChartConfig)

        if in_ipynb():
            iframe_src = 'http://localhost:3000/simulator/chart/'+get_current_kernel(self.workerJobs.workerId)

            self.__displayCoinChartFrame(iframe_src)

        return True


import time
import grpc
import coinlib.dataWorker_pb2 as statsModel
import coinlib.dataWorker_pb2_grpc as stats
import datetime 
import threading
import pandas as pd
import queue
import inspect
import numpy as np
import grpc
import asyncio

from coinlib.data.DataTable import DataTable
from coinlib.helper import log
import simplejson as json
import traceback
from coinlib.WorkerJobProcess import WorkerJobProcess
from coinlib.Registrar import Registrar
from multiprocessing import Pool
import multiprocessing
import zlib
from chipmunkdb.ChipmunkDb import ChipmunkDb


from coinlib.ChartsIndicatorJob import ChartsIndicatorJob

class ChartsSeriesDataResult:
    """ Point class represents and manipulates x,y coords. """

    def __init__(self):
        """ Create a new point at the origin """
        self.output_name = None
        self.result_data : DataTable = None
        self.indicator = None
        self.options = None


class ChartsWorker(WorkerJobProcess):
    
    def initialize(self):
        self.dataFrameLock = threading.Lock()
        self.registrar = Registrar()
        self.result_list = []
        self.final_result_list = []
        self.chartsInterface = stats.ChartsWorkerStub(self.getChannel())
        self.chartConfig = self.chartsInterface.GetConfig(self.workerJob)
        self.chartConfigData = self.chartConfig.chartData
        pass
        
    def getIndicatorConfiguration(self, element):
        return json.loads(str(element.indicatorConfig, 'ascii'))
    
    def getIndicatorMethod(self, indicator):
        found_method = None
        registeredFunctions = self.registrar.functionsCallbacks
        
        if registeredFunctions[indicator.feature+"."+indicator.subfeature] is not None:
            found_method = registeredFunctions[indicator.feature+"."+indicator.subfeature]
        
        return found_method

    def onErrorOnIndicatorHappened(self, indicator_element, error):

        indicatorError = statsModel.IndicatorError()
        indicatorError.error.message = str(error)
        indicatorError.worker.CopyFrom(self.workerJob)
        indicatorError.indicator.CopyFrom(indicator_element.indicator)

        self.chartsInterface.OnIndicatorErrorOccured(indicatorError)

        return False

    def runIndicatorCalculation(self, indicator_element, dataFrameCopy):
        process = None
        try:
            start = time.time()
            targetIndicatorFunction = self.getIndicatorMethod(indicator_element.indicator)
            inputs = self.getIndicatorConfiguration(indicator_element)

            raw_inputs = self.extractInputsFromDataFrameAndInsertInDataFrame(inputs)
            raw_inputs["name"] = indicator_element.indicator.subfeature
            raw_inputs["group"] = indicator_element.indicator.feature

            chart = ChartsIndicatorJob(indicator_element.indicator.subfeature, indicator_element.indicator.feature,
                                       inputs, dataFrameCopy, indicator_element, self)
            chart.setUniqueName(indicator_element.indicator.name)

            start = time.time()
            if (inspect.iscoroutinefunction(targetIndicatorFunction["process"])):
                async def runandwait():
                    try:
                        result = await asyncio.wait(
                            [targetIndicatorFunction["process"](raw_inputs, chart)],
                            return_when=asyncio.FIRST_EXCEPTION)
                    except Exception as e2:
                        raise e2

                loop = asyncio.new_event_loop()
                process = loop.run_until_complete(runandwait())
            else:
                process = targetIndicatorFunction["process"](raw_inputs, chart)

            log.info("running indicator finished " + indicator_element.indicator.subfeature, time.time() - start)
        except Exception as e:
            tb = traceback.format_exc()
            log.error(tb)
            log.error("Error on indicator")
            log.error(e)
            self.onErrorOnIndicatorHappened(indicator_element, e)
            pass

        return process

    def sendPartialChartData(self, indicator, output_name, options):

        partialData = statsModel.ChartsWorkerPartialDataLayout()
        partialData.chartInfo.name = output_name
        partialData.chartInfo.chartType = options["chartType"]
        if "color" in options:
            partialData.chartInfo.color = options["color"]
        if "size" in options:
            partialData.chartInfo.size = options["size"]
        if "opacity" in options:
            partialData.chartInfo.opacity = options["opacity"]

        # send raw data to Influx and then inform the listener
        #self.influxdb.insertChartSeries(output_name, result_data)

        indexes = self.dataFrame.index

       # self.influxDb.writeRawDataColumn(self.chartConfigData.chart_id, output_name, indexes, result_data)

        partialData.worker.CopyFrom(self.workerJob)
        partialData.indicator.CopyFrom(indicator.indicator)

        self.chartsInterface.OnPartialChartLayout(partialData)

    def pushAllResultsToCurrentDataFrame(self):

        targetColumns = []
        for result in self.result_list:

            targetLength = len(self.dataFrame.index)

            for key in result.result_data.getColumnNames():
                try:
                    result_rows = result.result_data.getColumn(key)
                    output_padded = np.pad(result_rows, (targetLength - len(result_rows), 0), 'constant',
                                           constant_values=np.nan)
                    self.dataFrame[result.output_name+key] = output_padded
                    targetColumns.append(result.output_name+key)
                except Exception as e:
                    pass



        return True

    def onPartialChartDataReceived(self, indicator, name, result_data, options):
        output_name = indicator.indicator.id_output

        name = name.replace(indicator.indicator.subfeature, "")

        if output_name is None or output_name == "":
            output_name = indicator.indicator.name + ("." + name if name != "" else "")

        ## lets send the data to the chart worker
        #t = threading.Thread(target=self.sendPartialChartData, args=[indicator, output_name, result_data, options], daemon=True)
        #t.start()
        self.sendPartialChartData(indicator, output_name, options)

        series = ChartsSeriesDataResult()
        series.output_name = output_name
        series.result_data = result_data
        series.indicator = indicator
        series.options = options
        self.result_list.append(series)

        for c in result_data.columns:
            self.final_result_list.append(output_name+c)

        return True
    
    def calculateNextElements(self, chartConfig):
        workerJobs = []
        self.result_list = []

        try:

            dataFrameCopy = self.dataFrame.copy()
            for element in chartConfig.elements:
                ##t = self.runIndicatorCalculation(element, return_dict)
                #t = multiprocessing.Process(target=self.runIndicatorCalculation, args=[element])
                t = threading.Thread(target=self.runIndicatorCalculation, args=[element, dataFrameCopy], daemon=True)

                workerJobs.append(t)

            # Start all threads
            for x in workerJobs:
                x.start()

            # Wait for all of them to finish
            for x in workerJobs:
                x.join()

            # pushing all received datas to the dataframe
            # and proceed to next corresponding indicators
            self.pushAllResultsToCurrentDataFrame()

        except Exception as ie:
            tb = traceback.format_exc()
            log.error(tb)
            log.error("Error on indicator")
            log.error(ie)
            self.onErrorOnIndicatorHappened(None, ie)

        try:
            if len(chartConfig.children.elements) > 0:
                self.calculateNextElements(chartConfig.children)
        except:
            pass
            
        return True
        
    def run(self):

        start = time.time()
        self.calculateNextElements(self.chartConfig)
        log.info("Finished complete calculation ", time.time() - start)
        ## let ssave all
        target_columns = self.final_result_list
        only_relevant_columns_df = self.dataFrame[target_columns]

        start = time.time()
        self.chipmunkDb.save_as_pandas(only_relevant_columns_df, self.chartConfigData.workspace_id,
                                  mode="append",
                                  domain=self.chartConfigData.chart_prefix)
        end = time.time()
        log.info("Saving DataFrame to chipmunk ", end - start)
        ##self.influxDb.writeDataFrame(self.chartConfigData.chart_id, self.dataFrame)
            
        


import numpy as np

from coinlib.BasicJobSessionStorage import BasicJobSessionStorage, LastSignalValue
from coinlib.data.DataTable import DataTable
from coinlib.helper import log
import pandas as pd

class BasicJob:
    def __init__(self, table: DataTable, inputs, storageManager = None):
        self.table = table
        if storageManager is None:
            storageManager = BasicJobSessionStorage()
        self._df = self.table.getDf()
        self.df = self.table.getDf()
        self._storageManager = storageManager
        self._storageManager.setCurrentIndex(self.date())
        self.inputs = inputs
        self.uniqueName = ""
        self._sessionInfo = self._storageManager.getStorage()
        self.features = self

        ## gets a signal data

    def set(self, name, data, index=None):

        return self.table.setColumn(name, data, index)

    def getOutputCol(self):
        return "result"

    def result(self, resultList, colname=None, fillType="front", type=np.float):

        #if isinstance(resultList, np.ndarray):
        #    resultList = np.pad(resultList, (self.df.shape[0] - len(resultList), 0), 'constant', constant_values=(np.nan))

        #self.df[self.getOutputCol()] = resultList
        self.table.setColumn(self.getOutputCol(), resultList, pad=True, type=type)

        return self.table.column(self.getOutputCol())

    ## This method adds a signal
    def wasSignal(self, name, maxSecondsDistance=0, minSecondsDistance=0, logicId=None):
        lastValue = self._storageManager.getLastSignal(logicId if logicId is not None else self.getUniqueName(), name)

        if lastValue.found is True:
            if minSecondsDistance > 0:
                if lastValue.distanceSeconds > minSecondsDistance:
                    return True
            if maxSecondsDistance > 0:
                if lastValue.distanceSeconds < maxSecondsDistance:
                    return True

        return False

    ## This method adds a signal
    def lastSignal(self, name, logicId=None) -> LastSignalValue:

        return self._storageManager.getLastSignal(logicId if logicId is not None else self.getUniqueName(), name)

    ## This method adds a signal
    def signal(self, name, data=None, index=-1, logicId=None) -> LastSignalValue:

        if data is not None:
            return self._storageManager.setSignal(logicId if logicId is not None else self.getUniqueName(), name, data)

        return self._storageManager.getSignal(logicId if logicId is not None else self.getUniqueName(), name)

    def getInputValue(self, input):

        if isinstance(input, dict):
            if "value" in input:
                return input["value"]

        return input


    def getNow(self, name):
        return self.current(self, name)

    def getCurrent(self, name):
        return self.current(self, name)

    def get(self, name, index=None, filterNone = False, replaceNone=None):

        data = None
        try:
                # if its a key of inputs - lets export the right column
                if name in self.inputs:
                    if isinstance(self.inputs[name], str):
                        return self.get(self.getInputValue(self.inputs[name]), index, filterNone, replaceNone)
                    if self.inputs[name]["type"] == "dataInput":
                        return self.get(self.getInputValue(self.inputs[name]["value"]), index, filterNone, replaceNone)

                data = None
                if name + ":y" in self.table.columns:
                    data = self.table.column(name + ":y")
                elif "additionalData."+name in self.table.columns:
                    data = self.table.column("additionalData."+name, index)
                elif name + ":close" in self.table.columns:
                    data = self.table.column(name + ":close", index)
                elif name in self.table.columns:
                    data = self.table.column(name, index)
                elif "stats."+name in self.table.columns:
                    data = self.table.column("stats."+name, index)
                elif name in self.inputs:
                    data = self.inputs[name]
        except Exception as e:
            log.error(e)

        if data is not None:
            if isinstance(data, list):
                if len(data) > 1:
                    data = [i if not np.isnan(i) else None for i in data]
                    if filterNone is True:
                        data = [i for i in data if i is not None]
                    if replaceNone is not None:
                        data = [0 if i is None else i for i in data]
                else:
                    return data[0]
            else:
                if np.isnan(data):
                    data = None


        return data

    def getAsArray(self, name, index=None, type=float):

        data = self.get(name, index=index)

        # filter data and remove nans in the beginning

        if len(data) > 0:
            if data[0] is None or np.isnan(data[0]):
                index = 0
                for f in data:
                    if f is None or np.isnan(f):
                        index = index + 1
                    else:
                        break
                data = data[index:]
        float_data = np.array(data, dtype=float)
        return float_data

    ## This method combines all params and combine as a dataframe
    def df(self):
        return self.df

    def current(self, name):
        return self.get(name, index=-1)

    def statistic(self, name="r_master"):

        return self.current("stats."+name)

    def setVar(self, name, data, logicId=None):

        return self.var(name, data, logicId=logicId)

    def setUniqueName(self, name):
        self.uniqueName = name
        return name

    def getUniqueName(self):
        return self.uniqueName

    def price(self, chart="chart1"):
        return self.table.getLast(chart+".main:close")

    def isNaN(self, num):
        return num != num

    def additional(self, name):

        index = -1

        if "additionalData." + name in self.table.columns:
            data = self.table.lastElement("additionalData." + name)
            if self.isNaN(data):
                return None
            return data

        return None

    def date(self):
        date = pd.to_datetime(self.table.index[-1])
        return date

    def time(self):
        date = pd.to_datetime(self.table.index[-1])
        return date

    ## This method adds a signal
    def var(self, name, data=None, logicId=None):

        if data is not None:
            return self._storageManager.setVar(logicId if logicId is not None else self.getUniqueName(), name, data)

        return self._storageManager.getVar(logicId if logicId is not None else self.getUniqueName(), name)




import asyncio

import grpc

from coinlib.brokerWorker.BrokerWorker import BrokerWorker
from coinlib.feature.FeatureWorker import FeatureWorkerWrapper
from coinlib.helper import is_in_ipynb
import coinlib.dataWorker_pb2 as statsModel
import coinlib.dataWorker_pb2_grpc as stats
from coinlib.helper import get_current_kernel
from coinlib.helper import log
from coolname import generate_slug
from coinlib.Registrar import Registrar
import time
import sys
from chipmunkdb.ChipmunkDb import ChipmunkDb

from coinlib.helper import pip_install_or_ignore
from coinlib.logics.LogicOfflineWorker import LogicOfflineWorker
from coinlib.logics.LogicOnlineWorker import LogicOnlineWorker
from coinlib.logics.LogicTestBrokerWorker import LogicTestBrokerWorker
from coinlib.notification.NotificationWorker import NotificationWorker
from coinlib.symbols.SymbolWorker import SymbolWorkerFetchHistoricalData, SymbolWorkerBrokerInfo, \
    SymbolWorkerConsumeMarketData

pip_install_or_ignore("semver", "semver")

import os

from coinlib.statistics import StatisticsMethodWorker, StatisticsRuleWorker
from coinlib import ChartsWorker
from coinlib import PluginsWorker

import threading
import queue

import semver
from grpc._cython.cygrpc import CompressionAlgorithm
from grpc._cython.cygrpc import CompressionLevel


class WorkerJobListener:
    channel = None
    def __init__(self, simulator=False):
        self.registrar = Registrar()
        self.waitingQueue = None
        self.workerSettings = None
        self.serverPlugins = None
        self.stopped = True
        self.simulator = simulator
        self.workerRegistered = False
        self.workerList = []

        file_name = os.path.basename(sys.argv[0]) + "-".join(self.registrar.worker_modules)
        fixed_worker_file = "."+file_name+".pid"
        if not is_in_ipynb:
            if os.path.exists(fixed_worker_file):
                with open(fixed_worker_file, 'r') as file:
                    self.workerId = file.read()
            else:
                self.workerId = self.generate_worker_id()
                f = open(fixed_worker_file, "a")
                f.write(self.workerId)
                f.close()
        else:
            self.workerId = self.generate_worker_id()

        availablePlugins = []
        try:
            all_plugins = self.getAllPluginsAvailable()
            for p in all_plugins:
                availablePlugins.append(statsModel.WorkerAvailablePlugins(name=p["name"], version=p["version"], type=p["module"]))
        except Exception as e:
            pass

        worker_modules = self.registrar.worker_modules
        if self.registrar.fixed_modules is not None:
            worker_modules = self.registrar.fixed_modules
        self.registrar.worker_id = self.workerId
        self.worker = statsModel.Worker(workerId=self.workerId,
                                        activeModules=worker_modules,
                                        availablePlugins=availablePlugins,
                                        workspaceId=self.registrar.workspaceId,
                                        environment=self.registrar.environment)
        return None

    def createRegistryChannnel(self):
        host = "localhost:9555"
        if is_in_ipynb:
            host = os.environ["COINLIB_WB"]+":9555"
        chan_ops = [('grpc.max_receive_message_length', 1000 * 1024 * 1024),
                    ('grpc.default_compression_algorithm', CompressionAlgorithm.gzip),
                    ('grpc.grpc.default_compression_level', CompressionLevel.high)]
        return grpc.insecure_channel(host, options=chan_ops,
                                     compression=grpc.Compression.Gzip)

    def createChannel(self):
        endpoint = self.registrar.workerEndpoint + ":" + self.registrar.workerEndpointPort
        chan_ops = [('grpc.max_receive_message_length', 1000 * 1024 * 1024),
                    ('metadata', "asdasd:asdasd"),
                    ('grpc.default_compression_algorithm', CompressionAlgorithm.gzip),
                    ('grpc.grpc.default_compression_level', CompressionLevel.high)]
        return grpc.insecure_channel(endpoint, options=chan_ops,

                                     compression=grpc.Compression.Gzip)

    def getChannel(self):
        return self.channel

    def onWorkerJobProcessError(self, workerJobProcess, err):
        try:
            self.workerList.remove(workerJobProcess)
        except Exception as e:
            pass
        log.info("Errored Worker: " + str(workerJobProcess.workerJob))
        log.info("Now worker running: " + str(len(self.workerList)))
        workerError = statsModel.WorkerJobError()
        workerError.workerJobId = workerJobProcess.workerJob.workerJobId
        workerError.errorMessage = str(err)
        workerError.workerJobType = workerJobProcess.workerJob.workerJobType
        self.stub.ErroredWorkerJob(workerError)

    def onWorkerJobProcessFinished(self, workerJobProcess):
        try:
            self.workerList.remove(workerJobProcess)
        except Exception as e:
            pass
        log.info("Finished Worker: " + str(workerJobProcess.workerJob))
        log.info("Now worker running: " + str(len(self.workerList)))
        self.stub.FinishedWorkerJob(workerJobProcess.workerJob)

    def generate_worker_id(self):
        if is_in_ipynb:
            return get_current_kernel()
        return self.get_random_string()

    def get_random_string(self, length=3):
        return generate_slug(length)

    def stop(self):
        self.stopped = True
        self.waitingQueue.put(None)

    def canWorkOnJob(self, job):
        return True

    def onJobStarted(self):
        pass

    def installPlugin(self, plugin):
        pluginWorker = PluginsWorker.PluginsWorker(None, self)
        pluginWorker.installPlugin(plugin)

    def checkIfPluginShouldInstall(self, pluginConfig):
        module_directory = ".chart_modules"
        if pluginConfig.type == "chart":
            module_directory = ".chart_modules"
        if pluginConfig.type == "notification":
            module_directory = ".notification_modules"
        elif pluginConfig.type == "stats":
            module_directory = ".statsrules_modules"
        elif pluginConfig.type == "symbols":
            module_directory = ".symbolbroker_modules"
        elif pluginConfig.type == "statsMethod":
            module_directory = ".statsmethod_modules"
        elif pluginConfig.type.endswith("Logics"):
            module_directory = ".logic_modules"
        elif pluginConfig.type == "symbolBroker":
            module_directory = ".symbolbroker_modules"
        elif pluginConfig.type == "broker":
            module_directory = ".broker_modules"
        elif pluginConfig.type == "features" or pluginConfig.type == "feature":
            module_directory = ".feature_modules"

        if not os.path.exists(module_directory):
            os.makedirs(module_directory)

        file_name = None
        version = "?"
        for file_name in os.listdir("./"+module_directory):
            try:
                if file_name.startswith(pluginConfig.name):
                    version = os.path.splitext(file_name)[0].split("_")[1]
                    version = version.replace("-", ".")
                    if semver.compare(pluginConfig.version, version) == 0:
                        return False
                    break
            except Exception as e:
                pass

        if file_name is not None:
            log.info("New Version for "+file_name+" => old: "+version+" new: "+pluginConfig.version)
        return True

    def getAllPluginsAvailable(self):

        plugins = []
        for module_directory in os.listdir("./"):
            if module_directory.startswith(".") and module_directory.endswith("_modules"):
                for file_name in os.listdir("./" + module_directory):
                    if file_name.endswith(".py"):
                        version = os.path.splitext(file_name)[0].split("_")[1]
                        plugin_name = os.path.splitext(file_name)[0].replace("_"+version, "")
                        version = version.replace("-", ".")
                        plugins.append({
                            "module": module_directory,
                            "name": plugin_name,
                            "file": file_name,
                            "file_path": "./"+module_directory+"/"+file_name,
                            "version": version
                        })

        return plugins

    def getDirectoryForPluginType(self, type):
        if type == "stats":
            return ".statsrules_modules"
        if type == "statsMethod":
            return ".statsmethod_modules"
        if type == "symbols":
            return ".symbolbroker_modules"
        if type == "broker":
            return ".broker_modules"
        if type == "notification":
            return ".notification_modules"
        if type == "chart":
            return ".chart_modules"
        if type == "symbolBroker":
            return ".symbolbroker_modules"
        if type == "features" or type == "feature":
            return ".feature_modules"
        if type.endswith("Logics"):
            return ".logic_modules"
        return None

    def downloadAllPlugins(self):
        try:
            self.serverPlugins = self.stub.GetAllPlugins(self.worker)

            foundPlugins = []
            for p in self.serverPlugins.plugin:
                try:
                    foundPlugins.append(p)
                    if self.checkIfPluginShouldInstall(p):
                        self.installPlugin(p)
                except Exception as e:
                    log.error("Error installing Plugin "+p.name, e)
                    pass

            ## remove all
            availablePlugins = self.getAllPluginsAvailable()

            for p in availablePlugins:
                found = False
                for p2 in foundPlugins:
                    dirName = self.getDirectoryForPluginType(p2.type)
                    if p["name"].lower() == p2.name.lower() and dirName == p["module"]:
                        found = True

                if found == False:
                    ## delete the plugin
                    #self.deletePlugin(p["file_path"])
                    log.info("Deleteing Plugin "+p["name"])
        except Exception as e:
            log.error("Error in download Plugins", e)

    def deletePlugin(self, full_path):
        return os.remove(full_path)

    def wait_for_jobs(self):

        try:
            self.workerSettings = self.stub.RegisterWorker(self.worker)
            self.updateWorkerSettings(self.workerSettings)
            self.downloadAllPlugins()
            self.workerRegistered = True

            self.workerStream =  self.stub.WatchWorkerJobs(iter(self.waitingQueue.get, None))
            m = statsModel.WorkerRegistration()
            m.worker_id = self.workerId
            self.waitingQueue.put(m)

            for job in self.workerStream:
                if (self.canWorkOnJob(job)):
                    self.acceptJob(job)
                else:
                    self.declineJob(job)

        except Exception as e:
            self.onErrorHappenedInCommunication(e)

    def onErrorHappenedInCommunication(self, e):
        start_time = threading.Timer(1, self.restart)
        start_time.start()

    def declineJob(self, job):
        self.stub.DeclineWorkerJob(job)

    def acceptJob(self, job):
        self.stub.AcceptWorkerJob(job)
        workerStartingThread = threading.Thread(target=self.startWorkerProcess, args=[job], daemon=True)
        workerStartingThread.start()

    def getDataWorkerEndpoint(self):
        res = statsModel.WorkerRegistryResponse()
        try:
            workerRequest = statsModel.WorkerRegistryRequest()
            workerRequest.worker_id = self.registrar.worker_id
            for m in self.registrar.worker_modules:
                workerRequest.activeModules.append(m)

            channel = self.createRegistryChannnel()
            registryStub = stats.DataWorkerRegistryStub(channel)
            res = registryStub.getDataWorkerServerEndpoint(workerRequest)
            self.registrar.workerEndpoint = res.targetendpoint
            self.registrar.workerEndpointPort = res.targetport
        except Exception as e:
            log.info("Can not connect to registry, i will retry", str(e))

    def connect(self):
        if self.channel is not None:
            try:
                self.channel.close()
            except Exception as e:
                pass

        self.getDataWorkerEndpoint()
        self.channel = self.createChannel()
        self.stub = stats.DataWorkerStub(self.channel)

    def getWorkerBulkJobData(self, chartData):
        start = time.time()
        self.chipmunkDb = ChipmunkDb(self.registrar.chipmunkdb)
        df = self.chipmunkDb.collection_as_pandas_additional(chartData.workspace_id, additionalCollections=[chartData.workspace_id+"_"+chartData.activity_id] ,columns=[])
        end = time.time()
        log.info("Downloading DataFrame from chipmunk ", end - start)

        return df

    def runBulkProcesses(self, workerJob):
        # lets download
        statisticInterface = stats.StatisticsMethodWorkerStub(self.getChannel())
        statisticConfig = statisticInterface.GetConfig(workerJob)

        # download
        dataFrame = self.getWorkerBulkJobData(statisticConfig.chartData)

        for workerJobChildConfig in statisticConfig.configs:
            try:
                workerJobChild = None
                if workerJobChildConfig.HasField("rule"):
                    workerJobChild = StatisticsRuleWorker.StatisticsRuleWorker(workerJob, self)
                    workerJobChild.setConfig(workerJobChildConfig.rule)
                else:
                    workerJobChild = StatisticsMethodWorker.StatisticsMethodWorker(workerJob, self)
                    workerJobChild.setConfig(workerJobChildConfig.methodWindow)

                if workerJobChild is not None:
                    workerJobChild.startProcessWithDataFrame(dataFrame.copy())

            except Exception as e:
                log.error("Data", str(e))
                pass


        return None

    def reRegisterWorker(self):
        self.connect()
        self.workerSettings = self.stub.RegisterWorker(self.worker)
        self.updateWorkerSettings(self.workerSettings)

    def updateWorkerSettings(self, settings: statsModel.WorkerSettings):

        activeModules = settings.activeModules

        for mod in activeModules:
            try:
                if mod == "chart":
                    self.registrar.functions.connect()
                    self.registrar.chartsFactory.loadPlugins()
                elif mod == "data":
                    self.registrar.data.connect()
                elif mod == "statsMethod":
                    self.registrar.statistics.connect()
                    self.registrar.statsMethodFactory.loadPlugins()
                elif mod == "broker":
                    self.registrar.broker.connect()
                    if self.registrar.isLiveEnvironment():
                        self.registrar.brokerFactory.loadPlugins()
                elif mod == "stats":
                    self.registrar.statistics.connect()
                    self.registrar.statsRuleFactory.loadPlugins()
                elif mod == "onlineLogics":
                    self.registrar.logic.connect()
                    self.registrar.logicFactory.loadPlugins()
                    self.registrar.logicFactory.loadChildModules(mod)
                elif mod.endswith("Logics"):
                    self.registrar.logic.connect()
                    self.registrar.logicFactory.loadPlugins()
                elif mod == "symbolBroker":
                    self.registrar.symbols.connect()
                    if self.registrar.isLiveEnvironment():
                        self.registrar.symbolFactory.loadPlugins()
                elif mod == "notification":
                    self.registrar.notifications.connect()
                    self.registrar.notificationFactory.loadPlugins()
                elif mod == "broker":
                    self.registrar.brokers.connect()
                    self.registrar.brokerFactory.loadPlugins()
                elif mod == "features" or mod == "feature":
                    self.registrar.features.connect()
                    self.registrar.featureFactory.loadPlugins()
                elif mod == "featureSaver":
                    self.registrar.featureSaverServer.start()
            except Exception as e:
                log.error("Error loading module "+mod, e)

        return None

    def generateWorkerProcess(self, workerJob):

       try:
           workerJobProcess = None
           # instantiate a new channel and stub so that the requests are parallel
           if workerJob.workerJobType == "stats":
               workerJobProcess = StatisticsRuleWorker.StatisticsRuleWorker(workerJob, self)
           elif workerJob.workerJobType == "statsMethod":
               workerJobProcess = StatisticsMethodWorker.StatisticsMethodWorker(workerJob, self)
           elif workerJob.workerJobType == "charts":
               workerJobProcess = ChartsWorker.ChartsWorker(workerJob, self)
           elif workerJob.workerJobType == "plugin":
               workerJobProcess = PluginsWorker.PluginsWorker(workerJob, self)
           elif workerJob.workerJobType == "fetchHistoricalData":
               workerJobProcess = SymbolWorkerFetchHistoricalData(workerJob, self)
           elif workerJob.workerJobType == "symbolBrokerInfo":
               workerJobProcess = SymbolWorkerBrokerInfo(workerJob, self)
           elif workerJob.workerJobType == "consumeMarketData":
               workerJobProcess = SymbolWorkerConsumeMarketData(workerJob, self)
           elif workerJob.workerJobType == "runLogicOffline":
               workerJobProcess = LogicOfflineWorker(workerJob, self)
           elif workerJob.workerJobType == "startBrokerSession":
               workerJobProcess = BrokerWorker(workerJob, self)
           elif workerJob.workerJobType == "testBrokerModule":
               workerJobProcess = LogicTestBrokerWorker(workerJob, self)
           elif workerJob.workerJobType == "startTraderSession":
               workerJobProcess = LogicOnlineWorker(workerJob, self)
           elif workerJob.workerJobType == "startFeature":
               workerJobProcess = FeatureWorkerWrapper(workerJob, self)
           elif workerJob.workerJobType == "sendNotification":
               workerJobProcess = NotificationWorker(workerJob, self)
           elif workerJob.workerJobType == "extractMessageCallback":
               workerJobProcess = NotificationWorker(workerJob, self)
           elif workerJob.workerJobType == "forcePluginUpdates":
               self.reloadAllPlugins()
           elif workerJob.workerJobType == "workerSettings":
               self.reRegisterWorker()
           else:
               log.error("Unknown Worker Type received")
               return workerJobProcess
       except Exception as e:
           log.error("Erorr in creating worker", e)
           pass

       return workerJobProcess


    def startWorkerProcess(self, workerJob):
        try:
            workerJobProcess = None

            if workerJob.workerJobType == "statsBulked":
                return self.runBulkProcesses(workerJob)
            else:
                workerJobProcess = self.generateWorkerProcess(workerJob)

            self.workerList.append(workerJobProcess)
            log.info("Now worker running: " + str(len(self.workerList)))

            workerJobProcess.startProcess()
            return workerJobProcess
        except Exception as e:
            log.error(e)
            pass

    def reloadAllPlugins(self):

        for mod in self.workerSettings.activeModules:
            try:
                self.reloadPluginsForType(mod)
            except Exception as e:
                pass
        return True

    def reloadPluginsForType(self, mod):

        if mod == "chart":
            self.registrar.chartsFactory.reloadPlugins()
        if mod == "statsMethod":
            self.registrar.statsMethodFactory.reloadPlugins()
        if mod == "stats":
            self.registrar.statsRuleFactory.reloadPlugins()
        if mod.endswith("Logics") or mod == "logic":
            self.registrar.logicFactory.reloadPlugins()
        if mod == "features" or mod == "feature":
            self.registrar.featureFactory.reloadPlugins()
        if mod == "broker":
            self.registrar.brokerFactory.reloadPlugins()
        if mod == "symbols" or mod == "symbolBroker":
            self.registrar.symbolFactory.reloadPlugins()
        if mod == "orderBroker" or mod == "broker":
            self.registrar.brokerFactory.reloadPlugins()

        return True

    def restart(self):
        if (self.stopped == False):
            self.connect()

            if (self.waitingQueue != None):
                self.waitingQueue.put(None)
            self.waitingQueue = None
            self.waitingQueue = queue.SimpleQueue()
            listenThread = threading.Thread(target=self.wait_for_jobs, daemon=True)
            listenThread.start()

    async def lockMethodUntilCallback(self):
        while True:
            if self.workerRegistered:
                return True

            await asyncio.sleep(0.1)
        return False

    def waitUntilWorkerRegistered(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.lockMethodUntilCallback())

    def start(self):
        if (self.stopped == False):
            self.stop()
        self.stopped = False
        self.connect()
        self.restart()
        self.waitUntilWorkerRegistered()





from coinlib.helper import in_ipynb, set_loglevel
from coinlib.helper import pip_install_or_ignore
from coinlib.helper import log, debug
import asyncio

from coinlib.featureSaver.FeatureSaverServer import FeatureSaverServer
from coinlib.notification.Notification import Notification
from coinlib.symbols.Symbols import Symbols

pip_install_or_ignore("ipynb_path", "ipynb_path")
# pip_install_or_ignore("import_ipynb", "import_ipynb")
pip_install_or_ignore("websocket", "websocket-client")
pip_install_or_ignore("plotly", "plotly")
pip_install_or_ignore("simplejson", "simplejson")
pip_install_or_ignore("asyncio", "asyncio")
pip_install_or_ignore("grpc", "grpcio-tools")
pip_install_or_ignore("matplotlib", "matplotlib")
pip_install_or_ignore("zlib", "zlib")
pip_install_or_ignore("pandas", "pandas")
pip_install_or_ignore("timeit", "timeit")
pip_install_or_ignore("dateutil", "python-dateutil")
pip_install_or_ignore("chipmunkdb", "chipmunkdb-python-client")

global coinlib_job_task

from coinlib.feature.Features import Features
from coinlib.Registrar import Registrar
from coinlib.logics.Logic import Logic
from coinlib.Functions import Functions
from coinlib.CoinlibDataInterface import CoinlibDataInterface
from coinlib.statistics.Statistics import Statistics
from coinlib.Simulator import Simulator
from coinlib.ChartsIndicatorJob import ChartsIndicatorJob
from coinlib.broker.Broker import Broker

registrar = Registrar()

if registrar.features is None:
    registrar.features = Features()
    registrar.logic = Logic()
    registrar.functions = Functions()
    registrar.statistics = Statistics()
    registrar.simulator = Simulator()
    registrar.symbols = Symbols()
    registrar.data = CoinlibDataInterface()
    registrar.featureSaverServer = FeatureSaverServer()
    registrar.notifications = Notification()
    registrar.broker = Broker()


features = registrar.features
logic = registrar.logic
functions = registrar.functions
data = registrar.data
notifications = registrar.notifications
symbols = registrar.symbols
statistics = registrar.statistics
simulator = registrar.simulator
broker = registrar.broker
pip_install_or_ignore = pip_install_or_ignore
set_loglevel = set_loglevel
debug = debug

def registerToWorkspace(workspaceId):
    registrar.workspaceId = workspaceId

def addLogicToWorkspace(identifier, type, logicComponentId, params=None, workspaceId=registrar.workspaceId):
    return logic.addLogicToWorkspace(identifier,
                                     type,
                                     logicComponentId,
                                     params, workspaceId)

def connectAsLogic(hostname=None, workspaceId=None):
    logicModule = "offlineLogics"
    if registrar.isLiveEnvironment():
        logicModule = "onlineLogics"
    return connectCoinlib(hostname, workspaceId=workspaceId, worker_modules=[logicModule])

def connectAsBrokerSymbol(hostname=None, workspaceId=None):
    return connectCoinlib(hostname, workspaceId=workspaceId, worker_modules=["symbolBroker"])

def connectAsBroker(hostname=None, workspaceId=None):
    return connectCoinlib(hostname, workspaceId=workspaceId, worker_modules=["broker"])

def connectAsFeature(hostname=None):
    return connectCoinlib(hostname, workspaceId=None, worker_modules=["features"])

def connectCoinlib(hostname=None, workspaceId=None, worker_modules=["data"], reconnect=False):
    if registrar.connected and not reconnect:
        return

    if not registrar.hasEnvironment():
        if hostname is not None:
            registrar.setBackendPath(hostname)
        registrar.setEnvironment("dev")
        registrar.fixed_modules = worker_modules
        registrar.worker_modules = worker_modules
        registrar.workspaceId = workspaceId

    registrar.connected = True
    registrar.simulator.connect()

def waitForJobs():

    # register as a worker if not already done
    # check if plugin and then ignore this
    if registrar.currentPluginLoading is None:

        loop = asyncio.get_event_loop()
        try:
            loop.run_forever()
        finally:
            loop.close()



import importlib
from IPython import get_ipython
from coinlib.helper import in_ipynb
from coinlib.helper import pip_install_or_ignore
import subprocess
import sys
import timeit
import math

#CHART_DATA_INFLUXDB_HOST="116.203.110.177"
#CHART_DATA_INFLUXDB_DB="coindeck-chartdata"
#CHART_DATA_INFLUXDB_USER="coindeck"
#CHART_DATA_INFLUXDB_PWD="6j8l0gbxsNkpAQXuL8C9"

CHART_DATA_INFLUXDB_HOST="localhost"
CHART_DATA_INFLUXDB_DB="coindeck-chartdata"
CHART_DATA_INFLUXDB_USER=""
CHART_DATA_INFLUXDB_PWD=""

pip_install_or_ignore("influxdb", "influxdb")

from influxdb import DataFrameClient
from influxdb import InfluxDBClient
from timeit import default_timer as timer
from datetime import timedelta



class InfluxDatabase:

    def __init__(self):
        self.database = CHART_DATA_INFLUXDB_DB
        self.influxClient = InfluxDBClient(CHART_DATA_INFLUXDB_HOST, 8086,
                                   CHART_DATA_INFLUXDB_USER, CHART_DATA_INFLUXDB_PWD, CHART_DATA_INFLUXDB_DB)
        self.dfclient = DataFrameClient(CHART_DATA_INFLUXDB_HOST, 8086,
                                   CHART_DATA_INFLUXDB_USER, CHART_DATA_INFLUXDB_PWD, CHART_DATA_INFLUXDB_DB)

    def getFullChartDataById(self, chart_id):
        data_result = self.dfclient.query('select * from "'+self.database+'"."autogen"."'+chart_id+'"')
        return data_result[chart_id]

    def writeDataFrameColumns(self, chart_id, dataFrame, columns):
        return self.dfclient.write_points(dataFrame, chart_id, field_columns=columns)


    def writeDataFrame(self, chart_id, dataFrame):
        return self.dfclient.write_points(dataFrame, chart_id)


    def writeRawDataColumn(self, chart_id, column, indexes, data):
        points = []
        data_length = 0
        column_formatter = ""
        columns = []
        type = "y"
        for col in data:
            columns.append(col)
            if col == "open":
                type = "ohlc"

        for key in data:
            data_length = len(data[key])
        for i in range(data_length):
            t = indexes[i]
            fields_data = {}
            if type == "y":
                if data["y"][i] is None or math.isnan(data["y"][i]):
                    continue
                fields_data[column+":y"] = data["y"][i]

            if type == "ohlc":
                if data["close"][i] is None or math.isnan(data["close"][i]):
                    continue
                fields_data[column + ":open"] = data["open"][i]
                fields_data[column + ":close"] = data["close"][i]
                fields_data[column + ":high"] = data["high"][i]
                fields_data[column + ":low"] = data["low"][i]
                fields_data[column + ":volume"] = data["volume"][i]
            row = {
                "measurement": chart_id,
                "time": t,
                "fields": fields_data
            }
            points.append(row)

        d = None
        if len(points) > 0:
            d = self.influxClient.write_points(points)
        return d


import asyncio
import calendar
import datetime
import os
import sys
from IPython.lib import kernel
from IPython import get_ipython
import importlib
import subprocess
import logging
import re

log = logging
currentfile = sys.argv[0]


def get_current_kernel(worker_id):
    if (is_in_ipynb == False):
        return worker_id

    connection_file_path = kernel.get_connection_file()
    connection_file = os.path.basename(connection_file_path)
    kernel_id = connection_file.split('-', 1)[1].split('.')[0]
    return kernel_id



class Serializable:
    def serialize(self, obj):
        """JSON serializer for objects not serializable by default json code"""

        return obj.__dict__

def serializeDTO(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime.date):
        return datetime.datetime.strftime(obj, "%Y-%m-%dT%H:%M:%S.%fZ")
    serialize_op = getattr(obj, "serialize", None)
    if callable(serialize_op):
        return obj.serialize(obj)
    if hasattr(obj, "__dict__"):
        return obj.__dict__

    return None

def in_ipynb():
    try:
        cfg = get_ipython().config

        if ("jupyter" in cfg['IPKernelApp']['connection_file']):
            return True
        elif ("Jupyter" in cfg['IPKernelApp']['connection_file']):
            return True
        else:
            return False
    except Exception:
        return False


is_in_ipynb = in_ipynb()
if is_in_ipynb:
    log.basicConfig(level=logging.ERROR)


def debug(debug=True):
    # Remove all handlers associated with the root logger object.
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)

    if debug:
        log.basicConfig(level=logging.INFO)
    else:
        log.basicConfig(level=logging.ERROR)
    return None


def set_loglevel(level):
    log.basicConfig(level=level)


if not 'workbookDir' in globals():
    workbookDir = os.getcwd()


def get_current_plugin_code_type():
    if is_in_ipynb:
        return "ipynb"
    return "py"

def checkIfCodeContainsVersionorExit( code):
    found = False
    try:
        found = re.search('(#version:)[ ]*([0-9a-z]*\.[0-9a-z]*\.[0-9a-z]*)([\n,\\\n])', code).group(2)
    except AttributeError:
        found = False

    if not found:
        print("ERROR: You should add '#version: 1.X.X' to your file")
        sys.exit(0)
    return found

def periodInSeconds(period):
    if period.endswith("m"):
        return float(period.replace("m", "")) * 60
    if period.endswith("h"):
        return float(period.replace("h", "")) * 60 * 60
    if period.endswith("d"):
        return float(period.replace("d", "")) * 60 * 60 * 24
    if period.endswith("M"):
        return float(period.replace("M", "")) * 60 * 60 * 24 * 30
    if period.endswith("Y"):
        return float(period.replace("Y", "")) * 60 * 60 * 24 * 365
    return 60


def isFinalTimestamp(coinlibtimeframe, date, lastFinalDate = None):
    mydate = None
    lastFinal = None
    if lastFinalDate is not None:
        if isinstance(lastFinalDate, str):
            lastFinal = datetime.datetime.strptime(lastFinalDate, "%Y-%m-%dT%H:%M:%S.%fZ")
        if isinstance(lastFinalDate, int):
            lastFinal = datetime.datetime.fromtimestamp(lastFinalDate / 1000)
        if isinstance(lastFinalDate, datetime.datetime):
            lastFinal = lastFinalDate
    if isinstance(date, str):
        mydate = datetime.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%fZ")
    if isinstance(date, int):
        mydate = datetime.datetime.fromtimestamp(date / 1000)
    if isinstance(date, datetime.datetime):
        mydate = date

    nowdate = datetime.datetime.now()

    if mydate is not None:
        if nowdate.weekday() == mydate.weekday() and \
                nowdate.month == mydate.month and \
                nowdate.year == mydate.year:
            timestamp = calendar.timegm(mydate.utctimetuple())
            periodSeconds = periodInSeconds(coinlibtimeframe)
            restofseconds = timestamp % periodSeconds
            if restofseconds == 0 and mydate.microsecond < 999000:
                if lastFinalDate is not None:
                    timestamp = calendar.timegm(mydate.replace(second=0, microsecond=0).utctimetuple())
                    lastFinaltimestamp = int(calendar.timegm(lastFinal.replace(second=0, microsecond=0).utctimetuple()))
                    if lastFinaltimestamp == timestamp:
                        return False
                print(mydate)
                return True

    return False


class Timer:
    def __init__(self, timeframe, coinlibClass, params, callback):
        self._periodSeconds = periodInSeconds(timeframe)
        self._params = params
        self._c = coinlibClass
        self._timeframe = timeframe
        self._callback = callback
        self._maximumInterval = self._periodSeconds / 60
        if self._maximumInterval > 60:
            self._maximumInterval = 60
        self._minimumInterval = 0.02
        self._dynamicInterval = 1
        self._is_first_call = True
        self._ok = True
        self._lastFired = 0
        self._task = asyncio.ensure_future(self._job())

    def checkTimeFrame(self, mydate):
        timestamp = calendar.timegm(mydate.utctimetuple())
        timestampSeconds = timestamp
        periodSeconds = periodInSeconds(self._timeframe)
        restofseconds = timestamp % periodSeconds
        if restofseconds == 0 \
                and self._lastFired < timestampSeconds:
            self._lastFired = int(timestamp)
            return True

        return self._periodSeconds - restofseconds

    async def _job(self):
        try:
            while self._ok:
                if self._dynamicInterval < self._minimumInterval:
                    self._dynamicInterval = self._minimumInterval
                if not self._is_first_call:
                    await asyncio.sleep(self._dynamicInterval)
                mydate = datetime.datetime.now()
                timeFrameSlot = self.checkTimeFrame(mydate)
                if isinstance(timeFrameSlot, bool) and timeFrameSlot == True:
                    await self._callback(self._c, self._timeframe, mydate, self._params)
                else:
                    if timeFrameSlot < 60:
                        self._dynamicInterval = 1
                    if timeFrameSlot < 10:
                        self._dynamicInterval = 0.5
                    if timeFrameSlot < 5:
                        self._dynamicInterval = 0.05
                    if timeFrameSlot >= 60:
                        self._dynamicInterval = self._maximumInterval
                self._is_first_call = False
        except Exception as ex:
            print(ex)

    def cancel(self):
        self._ok = False
        self._task.cancel()

def createTimerWithTimestamp(callback, timeframe, c, params):
    timer = Timer(timeframe=timeframe, coinlibClass=c, params=params, callback=callback)
    return timer


def find_current_runner_file(searchFor=[".waitForJobs", ".connectAsLogic",".connectAsBroker",
                                        ".connectAsFeature", ".connectAsNotification",
                                        ".connectCoinlib", ".connectAsBrokerSymbol"]):
    try:
        rel_path = ""
        last_path = ""
        for i in range(30):
            cur_path = sys._getframe(i).f_globals['__file__']
            code = get_current_plugin_code(cur_path)
            containsSearchFor = False
            try:
                for sf in searchFor:
                    if sf in code and sf + "\"" not in code:
                        containsSearchFor = True
            except AttributeError:
                containsSearchFor = False
                pass
            if containsSearchFor:
                rel_path = cur_path
                break
            last_path = cur_path

        cwd = os.getcwd()
        abs_path = os.path.join(cwd, rel_path)
        only_filename = os.path.splitext(os.path.basename(abs_path))[0]
        return abs_path, only_filename
    except Exception as e:
        printError("Can not find the code for your file - look if you connected coinlib Correct. (Use one of: "+",".join(searchFor)+" )")
        sys.exit(0)


def get_current_plugin_code(abs_path):
    if not is_in_ipynb:
        f = open(abs_path)
        return f.read()

    return ""

def printError(msg):
    print('\033[31m'+msg+'\033[31m')


def pip_install_or_ignore(import_name, module_name):
    try:
        return importlib.import_module(import_name)
    except ImportError:
        print("missing importing of " + import_name)
        if in_ipynb():
            get_ipython().run_line_magic('pip', 'install $module_name')
        else:
            subprocess.check_call([sys.executable, "-m", "pip", "install", module_name])
# Function to display hostname and



import os
import importlib
import pkgutil
import sys
import importlib
from pathlib import Path
from coinlib.helper import log
from coinlib import Registrar

class PluginLoader:

    def __init__(self, parentdirectory=""):
        self.registrar = Registrar.Registrar()
        self.parentDirectory = parentdirectory
        return None

    def loadChildModules(self, module):
        pass

    def hasPlugins(self):
        return True

    def loadPlugins(self):
        if not self.hasPlugins():
            return False
        dirname = self.parentDirectory + self.getLoaderPath()
        log.info("Loading plugins from "+ str(dirname))
        ##for importer, package_name, _ in pkgutil.iter_modules([dirname]):

        def onerror(name):
            print("Error importing module %s" % name)
            type, value, traceback = sys.exc_info()
            print(traceback)

        for loader, package_name, is_pkg in pkgutil.walk_packages([dirname], onerror=onerror):
            replaced_dirname = dirname.replace("/", ".")
            full_package_name = '%s.%s' % (replaced_dirname, package_name)
            if package_name not in sys.modules:
                log.info("Package is not in sys.modules "+package_name)
                log.info(dirname)
                self.loadPlugin(dirname+"/"+package_name)

        pass

    def reloadPlugins(self):
        if not self.hasPlugins():
            return False
        dirname = self.getLoaderPath()
        for importer, package_name, _ in pkgutil.iter_modules([dirname]):
            full_package_name = '%s.%s' % (dirname, package_name)
            log.info(full_package_name)
            self.reloadPlugin(dirname+"/"+package_name)

        return True

    def reloadPlugin(self, path):
        ret = False
        try:
            try:
                del sys.modules[path]
            except:
                pass

            # ret = __import__('%s' % path, globals=globals())
            base = os.path.basename(path)
            dirname = path.split("/")[0]
            path_comp = Path(path)
            moduleName = os.path.splitext(base)[0]
            sys.path.insert(0, dirname)
            self.registrar.currentPluginLoading = moduleName
            ret = __import__('%s' % moduleName, globals=globals())
            log.info("Reimported plugin: "+moduleName)
            self.registrar.currentPluginLoading = None
        except Exception as e:
            self.registrar.currentPluginLoading = None
            log.warning("ERror loading", str(e))
            pass
        return ret
        ##mod = importlib.import_module('%s' % path)
        ##return mod

    def loadPlugin(self, path):
        ret = None
        try:
            #ret = __import__('%s' % path, globals=globals())

            base = os.path.basename(path)
            dirname = path.split("/")[0]
            path_comp = Path(path)
            moduleName = os.path.splitext(base)[0]
            sys.path.insert(0, dirname)
            self.registrar.currentPluginLoading = moduleName
            ret = __import__('%s' % moduleName, globals=globals())
            log.info("Imported plugin: "+moduleName)
            self.registrar.currentPluginLoading = None

            #with open(path+".py", 'rb') as fp:
            #    globals()[moduleName] = importlib.import_module(moduleName, fp)
        except Exception as e:
            self.registrar.currentPluginLoading = None
            log.warning("Error loading", e)
            pass
        return ret
        ##mod = importlib.import_module('%s' % path)
        ##return mod









from coinlib.PluginLoader import PluginLoader

class ChartsFactory(PluginLoader):

    def __init__(self, parentdirectory=""):
        super(ChartsFactory, self).__init__(parentdirectory=parentdirectory)
        return None
        
    def getLoaderPath(self):
        return ".chart_modules"
    
    
    
    
    


import sys

from coinlib.brokerWorker.BrokerFactory import BrokerFactory
from coinlib.feature.FeatureFactory import FeatureFactory
from coinlib.helper import pip_install_or_ignore
from coinlib.notification.NotificationFactory import NotificationFactory
from coinlib.symbols import SymbolFactory

pip_install_or_ignore("ipynb_path", "ipynb_path")
pip_install_or_ignore("websocket", "websocket-client")
pip_install_or_ignore("plotly", "plotly")
pip_install_or_ignore("simplejson", "simplejson")
pip_install_or_ignore("asyncio", "asyncio")
pip_install_or_ignore("grpc", "grpcio-tools")
pip_install_or_ignore("matplotlib", "matplotlib")
pip_install_or_ignore("pandas", "pandas")
pip_install_or_ignore("timeit", "timeit")
pip_install_or_ignore("dateutil", "python-dateutil")
pip_install_or_ignore("chipmunkdb", "chipmunkdb-python-client")

from coinlib.logics import LogicLoader
from coinlib.statistics import StatisticsRuleFactory, StatisticsMethodFactory
from coinlib import ChartsFactory
from coinlib.DataWorker import WorkerJobListener
from coinlib.Registrar import Registrar
import asyncio
from coinlib.helper import log

registrar = Registrar()

def run_main_function(worker_modules = []):
    registrar.setEnvironment("live")
    registrar.worker_modules = worker_modules if worker_modules is not None else []
    registrar.statsRuleFactory = StatisticsRuleFactory.StatisticsRuleFactory()
    registrar.statsMethodFactory = StatisticsMethodFactory.StatisticsMethodFactory()
    registrar.chartsFactory = ChartsFactory.ChartsFactory()
    registrar.symbolFactory = SymbolFactory.SymbolFactory()
    registrar.notificationFactory = NotificationFactory()
    registrar.logicFactory = LogicLoader.LogicFactory()
    registrar.brokerFactory = BrokerFactory()
    registrar.featureFactory = FeatureFactory()


    workerJobs = WorkerJobListener()
    workerJobs.start()

    log.info("Starting Coinlib Factory")

    loop = asyncio.get_event_loop()
    try:
        loop.run_forever()
    finally:
        loop.close()

if __name__ == '__main__':
    worker_modules = []
    if len(sys.argv) > 1:
        worker_modules = sys.argv[1].split(",")
    run_main_function(worker_modules)


import os
from coinlib import log
from coinlib.ChartsFactory import ChartsFactory
from coinlib.brokerWorker.BrokerFactory import BrokerFactory
from coinlib.feature.FeatureFactory import FeatureFactory
from coinlib.logics.LogicLoader import LogicFactory
from coinlib.notification import NotificationFactory
from coinlib.statistics.StatisticsMethodFactory import StatisticsMethodFactory
from coinlib.statistics.StatisticsRuleFactory import StatisticsRuleFactory
from coinlib.symbols import SymbolFactory


class Registrar(object):
    _instance = None
    functionsCallbacks = {}
    statisticsCallbacks = {}
    logicCallbacks = {}
    featureCallbacks = {}
    symbolBrokerCallbacks = {}
    brokerCallbacks = {}
    notificationCallbacks = {}
    symbolFactory: SymbolFactory = None
    brokerFactory: BrokerFactory = None
    currentPluginLoading = None
    notificationFactory: NotificationFactory = None
    chartsFactory: ChartsFactory = None
    featureFactory: FeatureFactory = None
    workerEndpoint = None
    workerEndpointPort = None
    statsRuleFactory: StatisticsRuleFactory = None
    logicFactory: LogicFactory = None
    workspaceId = None
    environment = None
    statsMethodFactory: StatisticsMethodFactory = None
    worker_modules = []
    connected = False
    isRegistered = False
    iframe_host = None
    worker_id = None
    coinlib_backend = None
    chipmunkdb = "localhost"

    notifications = None
    simulator = None
    statistics = None
    logic = None
    brokers = None
    data = None
    functions = None
    features = None
    featureSaverServer = None
    fixed_modules = None


    def __new__(cls):
        if cls._instance is None:
            log.info('Creating the object')
            cls._instance = super(Registrar, cls).__new__(cls)
            # Put any initialization here.

        return cls._instance

    def hasEnvironment(self):
        return self.environment is not None

    def setEnvironment(self, env):
        if self.environment is None:
            self.environment = env
        else:
            log.error("You are trying to change the environment - thats probably an error?")

    def isLiveEnvironment(self):
        return self.environment == "live"
    
    def setBackendPath(self, path):
        self.iframe_host = path + ":3000"
        self.coinlib_backend = path + ":" + self.get_port()
       # self.chipmunkdb = path

    def get_port(self):
        return self.workerEndpointPort

    def set_coinlib_backend(self, endpoint):
        self.workerEndpoint = endpoint
        self.coinlib_backend = endpoint
        self.iframe_host = endpoint + ":3000"

    # IP address
    def get_coinlib_backend(self):
        return self.workerEndpoint

    def get_coinlib_backend_chipmunk(self):
        return self.get_coinlib_backend()

    def get_coinlib_backend_grpc(self):
        return self.get_coinlib_backend() + ":" + self.get_port()



import pandas as pd
import numpy as np
from coinlib.BasicJob import BasicJob
import time

from coinlib.data.DataTable import DataTable
from coinlib.helper import log

def isFullOHLC(row):
    return (row.close != 0 and row.close != None and row.open != 0 and row.open != None and
            row.high != 0 and row.high != None and row.low != None and row.low != 0)


class ChartsIndicatorJob(BasicJob):
    def __init__(self, name, group, inputs, df, indicator, worker):
        super(ChartsIndicatorJob, self).__init__(DataTable(df), inputs)
        self.name = name
        self.group = group
        self.send = False
        self.chunked = False
        self.worker = worker
        self.data = []
        self.indicator = indicator
    
    def isChunked(self):
        return self.chunked
    
    def setChunked(self, chunked):
        self.chunked = chunked
        return True

    def getUniqueName(self):
        return self.uniqeName

    def getDateTimeFromISO8601String(self, s):
        import datetime
        import dateutil.parser
        d = dateutil.parser.parse(s)
        return d

    
    def series(self, chartType, name, data, color=None, opacity=None, size=None):

        if (color == None):
            color = "#ccffcc"
        import datetime
        import pytz

        utc=pytz.UTC
        latestDate = datetime.datetime(1971, 3, 19, 13, 0, 9, 351812)
        latestDate = utc.localize(latestDate) 
        
        chart = {
            "chartType": chartType,
            "name": name,
            "data": [],
            "color": color
        }

        column_names = []
        # check if output data is a series, a dataframe or a array
        if isinstance(data, (list, pd.core.series.Series, pd.DataFrame, int, float, np.ndarray)):

            if isinstance(data, (list)):
                # its a list
                self.table.setColumn(":y", data)
                column_names = [":y"]

            elif isinstance(data, (int, float)):

                self.table.setColumn(":y", [data])
                column_names = [":y"]

            elif isinstance(data, (pd.core.series.Series)):

                self.table.setColumn(":y", data.values)
                column_names = [":y"]

            elif isinstance(data, (pd.DataFrame)):
                # its a dataframe
                self.table.setColumn(":open", data["open"])
                self.table.setColumn(":high", data["high"])
                self.table.setColumn(":low", data["low"])
                self.table.setColumn(":close", data["close"])
                self.table.setColumn(":volume",  data["volume"])
                column_names = [":open", ":high", ":low", ":close", ":volume"]

            elif isinstance(data, (np.ndarray)):
                # it is a numpy array
                self.table.setColumn(":y", data)
                column_names = [":y"]




        else:
            raise Exception("Please send your data as pandas.series, list, ndarray or pd.dataframe")

        options = {}

        if (color != None):
            options["color"] = color
        if (opacity != None):
            options["opacity"] = opacity
        if (size != None):
            options["size"] = size
        options["chartType"] = chartType
        """
        try:
            chart["data"] = df[column_names]
        except Exception as e:
            log.error(e)
            pass
        chart["options"] = options"""

        self.worker.onPartialChartDataReceived(self.indicator, name, self.table.subTable(columns=column_names), options)
        
        return True
    
    
    def line(self, chartType, name, selector, data):
        
        return True
    
    
    def plot(self, chartType, name, selector, data):
        
        return True
    
    def getSessionData(self):
        
        sess_list = None
        columns = []
        for colname in self.df.columns:
            if (colname.startswith("session:") and "Datetime" not in colname):
                columns.append(colname)
        
        if (len(columns) > 0):
            sessionDf = pd.DataFrame(self.df, columns=columns)

            sessionDf["date"] = sessionDf.index.to_series().apply(lambda x: x.isoformat())
            ###sessionDf.select_dtypes(exclude=['datetime', 'datetime64','datetimetz'])
            sessionDf.reset_index(drop=True, inplace=True)
            ##sessionDf.drop(columns='session:Datetime')

            sess_list = sessionDf.to_dict()
            
            

        return sess_list

    
        



from datetime import datetime

from chipmunkdb.ChipmunkDb import ChipmunkDb

import coinlib.dataWorker_pb2 as statsModel
import coinlib.dataWorker_pb2_grpc as stats
import simplejson as json
from coinlib.helper import is_in_ipynb
import ipynb_path
from grpc._cython.cygrpc import CompressionAlgorithm
from grpc._cython.cygrpc import CompressionLevel
from coinlib.Registrar import Registrar
import grpc

import inspect
from google.protobuf.json_format import MessageToDict
import os
import sys


class CoinlibDataInterface:
    channel = None

    def __init__(self):
        self.registrar = Registrar()


    def connect(self):
        self.channel = self.createChannel()
        self.dataInterface = stats.DataLoaderWorkerStub(self.channel)
        pass

    def createChannel(self):
        chan_ops = [('grpc.max_receive_message_length', 1000 * 1024 * 1024),
                    ('grpc.default_compression_algorithm', CompressionAlgorithm.gzip),
                    ('grpc.grpc.default_compression_level', CompressionLevel.high)]
        return grpc.insecure_channel(self.registrar.get_coinlib_backend_grpc(), options=chan_ops,
                                     compression=grpc.Compression.Gzip)

    def getAllExchanges(self):

        b = statsModel.ExchangeDataRequestInfo()

        ret = self.dataInterface.getAllExchanges(b)

        return ret.exchanges

    def getAllSymbols(self, additionalDataFeatures=[]):

        b = statsModel.SymbolDataRequestInfo()
        b.additionalDataFeatures = json.dumps(additionalDataFeatures)

        ret = self.dataInterface.getAllSymbols(b)

        return ret.symbols

    def getAvailableFeatures(self):

        b = statsModel.AdditionalDataFeatureRequest()

        ret = self.dataInterface.getAdditionalDataFeatures(b)

        return ret.featureInfos

    def getMarketData(self, symbol, startDate, endDate, timeframe="1h", exchange=None, contractType="spot", additionalDataFeatures=[]):

        b = statsModel.MarketDataRequest()
        s = statsModel.SymbolBrokerSymbolSmall()
        if exchange is not None:
            s.exchange_id = exchange
        s.symbol = symbol

        b.symbol.CopyFrom(s)
        b.timeframe = timeframe
        b.startDate = datetime.strftime(startDate, "%Y-%m-%dT%H:%M:%S.%fZ")
        b.endDate = datetime.strftime(endDate, "%Y-%m-%dT%H:%M:%S.%fZ")
        b.contractType = contractType
        b.additionalDataFeatures = json.dumps(additionalDataFeatures)

        ret = self.dataInterface.getMarketData(b)

        self.chipmunkDb = ChipmunkDb(self.registrar.chipmunkdb)
        df = self.chipmunkDb.collection_as_pandas_additional(ret.database_id, additionalCollections=[])

        return df



"""Setup script for coinlib"""
import os.path
from setuptools import setup

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README_pip.md")) as fid:
    README = fid.read()

# This call to setup() does all the work
setup(
    name="coinlib",
    version="1.2.0",
    description="Develop new code for your coindeck environment",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/coindeck/coinlib",
    author="coindeck",
    author_email="donnercody86@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],

    packages=["coinlib"],
    include_package_data=True,
    install_requires=[
        "requests", "semver", "datascience", "coolname", "google", "grpcio", "grpcio-tools", "protobuf", "cython", "ipython", "ipykernel", "pandas", "websocket-client", "plotly", "simplejson", "ipynb_path",
        "matplotlib", "pyarrow", "pandas", "python-dateutil", "chipmunkdb_python_client"
    ],
    entry_points={"console_scripts": ["coinlib=index:main"]},
)



import copy
import coinlib.dataWorker_pb2 as statsModel
import coinlib.dataWorker_pb2_grpc as stats
import json
from coinlib.helper import is_in_ipynb, get_current_plugin_code_type, find_current_runner_file, get_current_plugin_code
import ipynb_path
from grpc._cython.cygrpc import CompressionAlgorithm
from grpc._cython.cygrpc import CompressionLevel
import grpc
from coinlib.Registrar import Registrar
from google.protobuf.json_format import MessageToDict
import os
import sys

class Functions:
    def __init__(self):
        self.registrar = Registrar()
        pass

    def connect(self):
        self.channel = self.createChannel()
        self.functionsInterface = stats.FunctionsStub(self.channel)
        pass

    def createChannel(self):
        chan_ops = [('grpc.max_receive_message_length', 1000 * 1024 * 1024),
                    ('grpc.default_compression_algorithm', CompressionAlgorithm.gzip),
            ('grpc.grpc.default_compression_level', CompressionLevel.high)]
        return grpc.insecure_channel(self.registrar.get_coinlib_backend_grpc(), options=chan_ops, compression=grpc.Compression.Gzip)
        
    def registerChartFunction(self, process, group, name, short, inputs, description, 
                              dynamicTimeseries=False, unstablePeriod=False, mode=""):
        """
            dynamicTimeseries =  for example ATR based charts which have the special problem of "recalculating" past events (for example renko) 
                            If dynamicTimeseries != False then we will handle this indicator very special
        """
        inputs.insert(0, {"type": "symbol", "name": "symbol", "required": "true"})
        
        self.registerIndicatorFunction(process, group, name, short, 
                                       inputs, description, type="chartType", chartType="line", mode=mode,
                                       dynamicTimeseries=dynamicTimeseries, unstablePeriod=unstablePeriod)
        
        return True
        

    
    def registerIndicatorFunction(self, process, group, name, short, inputs, 
                                  description, chartType="", type="function", mode="",
                                  dynamicTimeseries=False, unstablePeriod=False):
        """Register Indicator with callback Functions.
        
            types: 
                - color
                - integer
                - float
                - symbol
                - series
                - select
                - feature
                    if you need some feature we fetch it for you and push it as an input
                    format of feature:
                
                    {name: inputname, feature: featureName, subfeature: None}
                
        
        """
        
        registration = statsModel.ChartWorkerIndicatorRegistration()
        registration.chartType = chartType
        registration.name = name
        registration.group = group
        registration.short_description = short
        registration.dynamicTimeseries = dynamicTimeseries
        registration.type = type
        registration.mode = mode
        registration.inputs = json.dumps(inputs, ensure_ascii=False).encode('gbk')
        registration.stage = self.registrar.environment
        registration.description = description


        if not is_in_ipynb:

            abs_path, only_filename = find_current_runner_file()
            registration.code = get_current_plugin_code(abs_path)

            if self.registrar.isLiveEnvironment():
                splitted = only_filename.split("_")
                only_filename_without_version = "_".join(splitted[:-1])
                registration.plugin = only_filename_without_version
                registration.pluginVersion = splitted[-1].replace("-", ".")
            else:
                registration.plugin = only_filename
                registration.pluginVersion = "?"
        else:
            abs_path = ipynb_path.get()
            only_filename = os.path.splitext(os.path.basename(abs_path))[0]

            registration.plugin = only_filename
            registration.pluginVersion = "?"

        registration.code_type = get_current_plugin_code_type()
        registration_dict = MessageToDict(registration)
        
        # before we send the data, we need to add the "callback"
        # to a global handler list
        self.registrar.functionsCallbacks[group+"."+name] = registration_dict
        self.registrar.functionsCallbacks[group+"."+name]["process"] = process
        
        self.functionsInterface.registerIndicatorFunction(registration)
        
        return True

    
